#include "wrap/thread.h"
#include "wrap/synch.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/**
 * Efter din förra lösning är du mycket nöjd eftersom programmet körde ungefär
 * dubbelt så snabbt som förut! Efter ytterligare lite funderande inser du att
 * det kanske finns flera saker som går att köra parallellt. Dock klarar din
 * förra lösning bara att starta en extra tråd, så du börjar med att skapa en
 * mer generell lösning på trådhanteringen så att det blir enkelt att starta
 * flera trådar samtidigt.
 *
 * För att lyckas med detta har du byggt vidare på din lösning och skapat två
 * funktioner som hjälper dig att hantera trådar: "spawn" och "wait".
 *
 * "spawn" skapar en tråd som kör "do_work" med den parametern som skickades
 * in. "spawn" returnerar en pekare till "struct work_data" som håller reda på
 * den skapade tråden.
 *
 * Den pekare som "spawn" returnerade kan sedan skickas till "wait" för att
 * vänta på att tråden blir klar, och för att få tillbaka resultatet.
 *
 * Du kan anta att "wait" bara anropas en gång för varje anrop till "spawn".
 */


// Funktion som gör de tunga beräkningarna. Denna vill vi köra parallellt i två
// trådar.
int do_work(int param) {
    // Här sker tungt arbete...
    // Tips, testa att avkommentera följande rad för att se problemen tydligare.
    // timer_msleep(param);

    // För enkelhets skull returnerar vi bara parametern i kvadrat.
    return param * param;
}

// Datastruktur som håller reda på en tråd som kör "do_work".
struct work_data {
    // Den parameter som ska skickas till "do_work".
    int param;

    // Resultatet som "do_work" har beräknat ifall tråden är klar.
    int result;
};

// Första funktionen som körs i nya trådar.
void thread_main(struct work_data *data) {
    data->result = do_work(data->param);
}

// Starta en ny tråd som kör funktionen "do_work" med "param" som
// parameter. Returnerar en "struct work_data" som sedan kan skickas till "wait"
// för att få reda på resultatet.
struct work_data *spawn(int param) {
    // Allokera en ny struktur för att hålla reda på tråden och initiera den.
    struct work_data *data = malloc(sizeof(struct work_data));
    data->param = param;

    // Skapa en ny tråd som kör "thread_main" och ge den tillgång till "data".
    thread_new(&thread_main, data);

    return data;
}

// Vänta på att en tråd som startades med "spawn" blir klar, och hämta
// resultatet från den. "wait" frigör också "data", så vi antar att "wait" bara
// anropas en gång för varje anrop till "spawn".
int wait(struct work_data *data) {
    // Hämta resultatet, frigör minnet och returnera resultatet.
    int result = data->result;
    free(data);
    return result;
}

// Main-funktion. Om implementationen ovan är korrekt ska du inte behöva ändra
// något här. Det kan dock vara intressant att modifiera "main" för att testa
// att din implementation fungerar som den ska.
int main(void) {
    struct work_data *a = spawn(10);
    struct work_data *b = spawn(100);

    int c = do_work(5);

    printf("Result for 'a': %d\n", wait(a));
    printf("Result for 'b': %d\n", wait(b));
    printf("Result for 'c': %d\n", c);

    return 0;
}
