#include "wrap/thread.h"
#include "wrap/synch.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

/**
 * Efter att ha läst på om ytterligare ideer gällande synkronisering hittar du
 * någon som föreslår att använda så kallade "pipelines" för att behandla data i
 * flera trådar. En pipeline är i princip flera trådar som var och en gör ett
 * litet steg av en beräkning, och sedan skickar sin data vidare till nästa steg
 * via en delad datastruktur. Här har vi implementerat den i en datatyp "buffer".
 *
 * Buffern är ett ställe som kan lagra lite data från en tråd i väntan på att
 * nästa tråd ska ta datan och behandla den vidare. För att förenkla
 * implementationen har buffern bara plats för ett element. Den har
 * fyra operationer:
 *
 * "buffer_create" - skapar en buffer.
 *
 * "buffer_destroy" - förstör en buffer.
 *
 * "buffer_put" - lägger ett element i buffern. Om buffern är full ska
 * implementationen vänta på att det finns plats i buffern innan insättning
 * sker.
 *
 * "buffer_get" - tar bort ett element från buffern. Om buffern är tom ska
 * implementationen vänta på att det finns ett element.
 */

// Datastrukturen.
struct buffer {
    // Det värde som finns i buffern.
    int value;
};

// Skapa en buffer.
struct buffer *buffer_create(void) {
    // Allokera minne och initiera.
    struct buffer *buffer = malloc(sizeof(struct buffer));
    buffer->value = 0;
    return buffer;
}

// Deallokera en buffer.
void buffer_destroy(struct buffer *buffer) {
    free(buffer);
}

// Lägg ett element i buffern. Om buffern är full ska vi vänta på att det finns
// plats innan insättning sker.
void buffer_put(struct buffer *buffer, int value) {
    buffer->value = value;
}

// Ta bort ett element från buffern. Om buffern är tom ska vi vänta på att det
// finns ett element.
int buffer_get(struct buffer *buffer) {
    return buffer->value;
}

/**
 * Huvudprogram. Om implementationen ovan är korrekt ska implementationen nedan
 * inte behöva ändras. Dock kan det vara intressant att göra ändringar för att
 * testa din implementation.
 */

// Tråd som producerar data till huvudtråden.
void thread(struct buffer *buffer) {
    for (int i = 0; i < 10; i++) {
        buffer_put(buffer, i);
    }
}

int main(void) {
    struct buffer *buffer = buffer_create();

    // Skapa en tråd som producerar data till oss.
    thread_new(&thread, buffer);

    // Hämta data.
    for (int i = 0; i < 10; i++) {
        printf("Got: %d\n", buffer_get(buffer));
    }

    // Städa.
    buffer_destroy(buffer);

    return 0;
}
